package impl

import (
	"context"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/log"
	"gitlab.okta-solutions.com/mashroom/backend/maintenance/api"
	"gitlab.okta-solutions.com/mashroom/backend/maintenance/model"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"time"
)

func (server *serverImpl) Create(ctx context.Context, request *maintenance.CreateRequest) (*maintenance.CreateResponse, error) {

	if request.User == "" {
		return nil, model.ErrNoUser
	}
	if request.Maintenance == nil || request.Maintenance.GetProperty() == nil || request.Maintenance.GetPropertyValue() == "" {
		return nil, model.ErrNoProperty
	}

	newMaintenance := &model.MaintenanceModel{
		Id:        primitive.NewObjectID(),
		CreatedAt: time.Now(),
		UpdatedAt: time.Now(),
		Status:    maintenance.MaintenanceStatus_NOT_ADDRESSED,
		User:      request.User,
	}

	err := newMaintenance.UpdateMaintenanceModel(request.GetMaintenance(), true)
	if err != nil {
		return nil, err
	}

	newMaintenance.UpdateLandlord(ctx)

	err = newMaintenance.ValidUser(ctx, request.User)
	if err != nil {
		return nil, err
	}

	err = model.InsertMaintenance(ctx, newMaintenance)
	if err != nil {
		return nil, err
	}

	err = newMaintenance.SendNotification(ctx, request.User, model.MaintenceEvent_CREATE)
	if err != nil {
		log.WithError(err).Error("sending maintenance create")
	}

	return &maintenance.CreateResponse{
		Success:     true,
		Maintenance: newMaintenance.RenderMaintenanceModel(),
	}, nil
}

func (server *serverImpl) Update(ctx context.Context, request *maintenance.UpdateRequest) (*maintenance.UpdateResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.GetMaintenance().GetIdValue())
	if err != nil {
		return nil, err
	}

	old, err := model.GetMaintenance(ctx, id)
	if err != nil {
		return nil, err
	}

	err = old.ValidUser(ctx, request.User)
	if err != nil {
		return nil, err
	}

	old.UpdatedAt = time.Now()
	err = old.UpdateMaintenanceModel(request.GetMaintenance(), false)
	if err != nil {
		return nil, err
	}

	err = model.UpdateMaintenance(ctx, old)
	if err != nil {
		return nil, err
	}

	event, err := old.GetUpdateType(ctx, request.User)
	if err != nil {
		return nil, err
	}

	err = old.SendNotification(ctx, request.User, event)
	if err != nil {
		log.WithError(err).Error("sending maintenance update")
	}

	return &maintenance.UpdateResponse{
		Success:     true,
		Maintenance: old.RenderMaintenanceModel(),
	}, nil
}

func (server *serverImpl) Get(ctx context.Context, request *maintenance.GetRequest) (*maintenance.GetResponse, error) {
	id, err := primitive.ObjectIDFromHex(request.Id)
	if err != nil {
		return nil, err
	}

	old, err := model.GetMaintenance(ctx, id)
	if err != nil {
		return nil, err
	}

	err = old.ValidUser(ctx, request.User)
	if err != nil {
		return nil, err
	}

	return &maintenance.GetResponse{
		Success:     true,
		Maintenance: old.RenderMaintenanceModel(),
	}, nil
}

func (server *serverImpl) List(ctx context.Context, request *maintenance.ListRequest) (*maintenance.ListResponse, error) {
	list, err := model.GetListMaintenance(ctx, request.User)
	if err != nil {
		return nil, err
	}

	maintenances := make([]*maintenance.Maintenance, 0)
	for _, mn := range list {
		maintenances = append(maintenances, mn.RenderMaintenanceModel())
	}
	return &maintenance.ListResponse{
		Success:     true,
		Maintenance: maintenances,
	}, nil
}

func (server *serverImpl) ListIncoming(ctx context.Context, request *maintenance.ListIncomingRequest) (*maintenance.ListIncomingResponse, error) {
	list, err := model.GetListIncomingMaintenance(ctx, request.User)
	if err != nil {
		return nil, err
	}

	maintenances := make([]*maintenance.Maintenance, 0)
	for _, mn := range list {
		maintenances = append(maintenances, mn.RenderMaintenanceModel())
	}
	return &maintenance.ListIncomingResponse{
		Success:     true,
		Maintenance: maintenances,
	}, nil
}

func (server *serverImpl) ListOutgoing(ctx context.Context, request *maintenance.ListOutgoingRequest) (*maintenance.ListOutgoingResponse, error) {
	list, err := model.GetListOutgoingMaintenance(ctx, request.User)
	if err != nil {
		return nil, err
	}

	maintenances := make([]*maintenance.Maintenance, 0)
	for _, mn := range list {
		maintenances = append(maintenances, mn.RenderMaintenanceModel())
	}
	return &maintenance.ListOutgoingResponse{
		Success:     true,
		Maintenance: maintenances,
	}, nil
}
