package model

import (
	"context"
	"errors"
	"github.com/golang/protobuf/ptypes/timestamp"
	"gitlab.okta-solutions.com/mashroom/backend/account/api"
	"gitlab.okta-solutions.com/mashroom/backend/common/api/generic"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/env"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/grpc"
	"gitlab.okta-solutions.com/mashroom/backend/common/generic/log"
	maintenance "gitlab.okta-solutions.com/mashroom/backend/maintenance/api"
	"gitlab.okta-solutions.com/mashroom/backend/messaging/api"
	"gitlab.okta-solutions.com/mashroom/backend/notification/api"
	"gitlab.okta-solutions.com/mashroom/backend/property/api"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"time"
)

var (
	PropertyServiceUrl = env.ResolveString("PROPERTY_SERVICE_URL", "mashroom-property:10000")
	PropertyService    = property.NewPropertyServiceClient(grpc.Dial(PropertyServiceUrl, "Property"))

	AccountServiceUrl = env.ResolveString("ACCOUNT_SERVICE_URL", "mashroom-account:10000")
	AccountService    = account.NewAccountServiceClient(grpc.Dial(AccountServiceUrl, "Account"))

	NotificationServiceUrl = env.ResolveString("NOTIFICATION_SERVICE_URL", "mashroom-notification:10000")
	NotificationService    = notification.NewNotificationServiceClient(grpc.Dial(NotificationServiceUrl, "Notification"))
)

var ErrNoUser = errors.New("user not found")
var ErrNoUserContact = errors.New("user contact not found")
var ErrNoMaintenance = errors.New("no maintenance found")
var ErrNoAccess = errors.New("no access")
var ErrNoProperty = errors.New("property not found")
var ErrNoNotificationEvent = errors.New("notification event not found")

type MaintenceEvent int32

const (
	MaintenceEvent_UNKNOWN         MaintenceEvent = 0
	MaintenceEvent_CREATE          MaintenceEvent = 1
	MaintenceEvent_UPDATE_LANDLORD MaintenceEvent = 2
	MaintenceEvent_UPDATE_TENANT   MaintenceEvent = 3
)

type SpecialistModel struct {
	FirstName   string    `bson:"first_name"`
	LastName    string    `bson:"last_name"`
	PhoneNumber string    `bson:"phone_number"`
	VisitAt     time.Time `bson:"visit_at"`
}

type MaintenanceModel struct {
	Id            primitive.ObjectID                `bson:"_id"`
	User          string                            `bson:"user"`
	Landlord      string                            `bson:"landlord"`
	Property      string                            `bson:"property"`
	Title         string                            `bson:"title"`
	Description   string                            `bson:"description"`
	Status        maintenance.MaintenanceStatus     `bson:"status"`
	Documents     []string                          `bson:"documents"`
	OccurringDate time.Time                         `bson:"occurring_date"`
	RepairDate    time.Time                         `bson:"repair_date"`
	CreatedAt     time.Time                         `bson:"created_at"`
	UpdatedAt     time.Time                         `bson:"updated_at"`
	Place         maintenance.MaintenancePlace      `bson:"place"`
	Category      maintenance.MaintenanceCategory   `bson:"category"`
	Issue         maintenance.MaintenanceIssueType  `bson:"issue"`
	Resolution    maintenance.MaintenanceResolution `bson:"resolution"`
	Specialist    *SpecialistModel                  `bson:"specialist"`
}

func FromUnixTime(t *timestamp.Timestamp) time.Time {
	if t == nil {
		return time.Time{}
	}
	return time.Unix(t.Seconds, int64(t.Nanos))
}

func ToUnixTime(t time.Time) *timestamp.Timestamp {
	return &timestamp.Timestamp{
		Seconds: t.Unix(),
		Nanos:   int32(t.Nanosecond()),
	}
}

func uniqueStrings(in []string) (out []string) {
	m := make(map[string]bool)
	for _, s := range in {
		if _, ok := m[s]; ok {
			continue
		}
		m[s] = true
		out = append(out, s)
	}
	return
}

func (model *MaintenanceModel) RenderMaintenanceModel() *maintenance.Maintenance {
	return &maintenance.Maintenance{
		Id:          &maintenance.Maintenance_IdValue{IdValue: model.Id.Hex()},
		User:        &maintenance.Maintenance_UserValue{UserValue: model.User},
		Property:    &maintenance.Maintenance_PropertyValue{PropertyValue: model.Property},
		Status:      &maintenance.Maintenance_StatusValue{StatusValue: model.Status},
		Title:       &maintenance.Maintenance_TitleValue{TitleValue: model.Title},
		Description: &maintenance.Maintenance_DescriptionValue{DescriptionValue: model.Description},
		Documents: &maintenance.Maintenance_DocumentsValue{
			DocumentsValue: &generic.StringArrayValue{Value: model.Documents},
		},
		CreatedAt: &maintenance.Maintenance_CreatedAtValue{CreatedAtValue: ToUnixTime(model.CreatedAt)},
		UpdatedAt: &maintenance.Maintenance_UpdatedAtValue{UpdatedAtValue: ToUnixTime(model.UpdatedAt)},

		OccurringDate: &maintenance.Maintenance_OccurringDateValue{OccurringDateValue: ToUnixTime(model.OccurringDate)},
		RepairDate:    &maintenance.Maintenance_RepairDateValue{RepairDateValue: ToUnixTime(model.RepairDate)},

		Place:      &maintenance.Maintenance_PlaceValue{PlaceValue: model.Place},
		Category:   &maintenance.Maintenance_CategoryValue{CategoryValue: model.Category},
		Issue:      &maintenance.Maintenance_IssueValue{IssueValue: model.Issue},
		Resolution: &maintenance.Maintenance_ResolutionValue{ResolutionValue: model.Resolution},
		Specialist: &maintenance.Maintenance_SpecialistValue{SpecialistValue: model.Specialist.RenderSpecialistModel()},
	}
}

func (model *SpecialistModel) RenderSpecialistModel() *maintenance.Specialist {
	if model == nil {
		return &maintenance.Specialist{}
	}

	return &maintenance.Specialist{
		FirstName:   &maintenance.Specialist_FirstNameValue{FirstNameValue: model.FirstName},
		LastName:    &maintenance.Specialist_LastNameValue{LastNameValue: model.LastName},
		PhoneNumber: &maintenance.Specialist_PhoneNumberValue{PhoneNumberValue: model.PhoneNumber},
		VisitAt:     &maintenance.Specialist_VisitAtValue{VisitAtValue: ToUnixTime(model.VisitAt)},
	}
}

func (model *MaintenanceModel) UpdateMaintenanceModel(mn *maintenance.Maintenance, create bool) error {
	if mn.GetId() != nil {
		if v, err := primitive.ObjectIDFromHex(mn.GetIdValue()); err != nil {
			return err
		} else {
			model.Id = v
		}
	}

	if create && mn.GetCreatedAt() != nil {
		model.CreatedAt = FromUnixTime(mn.GetCreatedAtValue())
	}
	if create && mn.GetUser() != nil {
		model.User = mn.GetUserValue()
	}
	if create && mn.GetProperty() != nil {
		model.Property = mn.GetPropertyValue()
	}

	if mn.GetUpdatedAt() != nil {
		model.UpdatedAt = FromUnixTime(mn.GetUpdatedAtValue())
	}
	if mn.GetOccurringDate() != nil {
		model.OccurringDate = FromUnixTime(mn.GetOccurringDateValue())
	}
	if mn.GetRepairDate() != nil {
		model.RepairDate = FromUnixTime(mn.GetRepairDateValue())
	}
	if mn.GetStatus() != nil {
		model.Status = mn.GetStatusValue()
	}
	if mn.GetTitle() != nil {
		model.Title = mn.GetTitleValue()
	}
	if mn.GetDescription() != nil {
		model.Description = mn.GetDescriptionValue()
	}
	if mn.GetDocuments() != nil {
		model.Documents = uniqueStrings(mn.GetDocumentsValue().GetValue())
	}

	if mn.GetPlace() != nil {
		model.Place = mn.GetPlaceValue()
	}
	if mn.GetCategory() != nil {
		model.Category = mn.GetCategoryValue()
	}
	if mn.GetIssue() != nil {
		model.Issue = mn.GetIssueValue()
	}
	if mn.GetResolution() != nil {
		model.Resolution = mn.GetResolutionValue()
	}
	if mn.GetSpecialist() != nil {
		model.Specialist = &SpecialistModel{
			FirstName:   mn.GetSpecialistValue().GetFirstNameValue(),
			LastName:    mn.GetSpecialistValue().GetLastNameValue(),
			PhoneNumber: mn.GetSpecialistValue().GetPhoneNumberValue(),
			VisitAt:     FromUnixTime(mn.GetSpecialistValue().GetVisitAtValue()),
		}
	}

	return nil
}

func (model *MaintenanceModel) ValidUser(ctx context.Context, user string) error {
	if user == "" {
		return ErrNoUser
	}

	if model.User == user {
		return nil
	}

	prop, err := PropertyService.Get(ctx, &property.GetRequest{
		Id: model.Property,
	})
	if err != nil {
		log.Error("no property", model.Property)
		return ErrNoProperty
	}

	if prop.GetProperty().GetPeople() != nil {
		people := prop.GetProperty().GetPeopleValue()
		if people.GetLandlordValue() == user {
			return nil
		}
		if people.GetTenants() != nil {
			tenants := people.GetTenantsValue().Value
			for _, v := range tenants {
				if v.GetUser() != nil {
					if v.GetUserValue() == user {
						return nil
					}
				}
			}
		}
	}
	return ErrNoAccess
}

func (model *MaintenanceModel) GetUpdateType(ctx context.Context, user string) (MaintenceEvent, error) {
	if user == "" {
		return MaintenceEvent_UNKNOWN, ErrNoUser
	}

	prop, err := PropertyService.Get(ctx, &property.GetRequest{
		Id: model.Property,
	})
	if err != nil {
		log.Error("no property", model.Property)
		return MaintenceEvent_UNKNOWN, ErrNoProperty
	}

	if prop.GetProperty().GetPeople() != nil {
		people := prop.GetProperty().GetPeopleValue()
		if people.GetLandlordValue() == user {
			return MaintenceEvent_UPDATE_TENANT, nil
		}
		if people.GetTenants() != nil {
			tenants := people.GetTenantsValue().Value
			for _, v := range tenants {
				if v.GetUser() != nil {
					if v.GetUserValue() == user {
						return MaintenceEvent_UPDATE_LANDLORD, nil
					}
				}
			}
		}
	}
	return MaintenceEvent_UNKNOWN, ErrNoUser
}

func (model *MaintenanceModel) SendNotification(ctx context.Context, user string, maintenceEvent MaintenceEvent) error {
	if user == "" {
		return ErrNoUser
	}

	// if model.User == user {
	// 	return nil
	// }

	prop, err := PropertyService.Get(ctx, &property.GetRequest{
		Id: model.Property,
	})
	if err != nil {
		log.Error("no property", model.Property)
		return ErrNoProperty
	}

	users := make([]string, 0)
	if prop.GetProperty().GetPeople() != nil {
		people := prop.GetProperty().GetPeopleValue()
		if people.GetLandlordValue() == user {
			// notify tenants
			if people.GetTenants() != nil {
				tenants := people.GetTenantsValue().Value
				for _, v := range tenants {
					if v.GetUser() != nil {
						users = append(users, v.GetUserValue())
					}
				}
			}
		} else {
			// notify landlord
			users = append(users, people.GetLandlordValue())
		}
	}

	if len(users) > 0 {
		var event notification.EventType
		switch maintenceEvent {
		case (MaintenceEvent_CREATE):
			event = notification.EventType_EVENT_MAINTENANCE_CREATE
		case (MaintenceEvent_UPDATE_LANDLORD):
			event = notification.EventType_EVENT_MAINTENANCE_UPDATE_LANDLORD
		case (MaintenceEvent_UPDATE_TENANT):
			event = notification.EventType_EVENT_MAINTENANCE_UPDATE_TENANT
		default:
			return ErrNoNotificationEvent
		}
		for _, recipient := range users {
			_, err = NotificationService.SendEvent(ctx, &notification.SendEventRequest{
				User:     recipient,
				Event:    event,
				Channels: []notification.EventChannel{notification.EventChannel_EVENT_CHANNEL_EMAIL},
				Embeds: []*notification.EventEmbed{
					{
						EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
						EmbedValue: model.Property,
					},
					{
						EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
						EmbedValue: user,
					},
				},
			})
		}
	}

	return ErrNoAccess
}

func (model *MaintenanceModel) UpdateLandlord(ctx context.Context) {
	if model.Property == "" {
		return
	}

	prop, err := PropertyService.Get(ctx, &property.GetRequest{
		Id: model.Property,
	})
	if err != nil {
		return
	}

	if prop.GetProperty().GetPeople() != nil {
		people := prop.GetProperty().GetPeopleValue()
		if people.GetLandlord() != nil {
			model.Landlord = people.GetLandlordValue()
		}
	}
}

func InsertMaintenance(ctx context.Context, prop *MaintenanceModel) error {
	prop.Id = primitive.NewObjectID()
	_, err := Maintenances.InsertOne(ctx, prop)
	return err
}

func GetMaintenance(ctx context.Context, id primitive.ObjectID) (*MaintenanceModel, error) {
	var result = &MaintenanceModel{}
	err := Maintenances.FindOne(ctx, bson.M{"_id": id}).Decode(result)
	return result, err
}

func UpdateMaintenance(ctx context.Context, mn *MaintenanceModel) error {
	log.Println(mn)
	_, err := Maintenances.ReplaceOne(ctx, bson.M{"_id": mn.Id}, mn)
	return err
}

func GetListMaintenance(ctx context.Context, user string) ([]*MaintenanceModel, error) {
	var result []*MaintenanceModel

	cur, err := Maintenances.Find(ctx, bson.M{"$or": bson.A{
		bson.M{"user": user},
		bson.M{"landlord": user},
	}})
	if err != nil {
		return nil, err
	}
	for cur.Next(ctx) {
		mn := &MaintenanceModel{}
		err = cur.Decode(mn)
		if err != nil {
			return nil, err
		}
		result = append(result, mn)
	}

	return result, err
}
func GetListIncomingMaintenance(ctx context.Context, user string) ([]*MaintenanceModel, error) {
	var result []*MaintenanceModel

	cur, err := Maintenances.Find(ctx, bson.M{"landlord": user})
	if err != nil {
		return nil, err
	}
	for cur.Next(ctx) {
		mn := &MaintenanceModel{}
		err = cur.Decode(mn)
		if err != nil {
			return nil, err
		}
		result = append(result, mn)
	}

	return result, err
}
func GetListOutgoingMaintenance(ctx context.Context, user string) ([]*MaintenanceModel, error) {
	var result []*MaintenanceModel

	cur, err := Maintenances.Find(ctx, bson.M{"user": user})
	if err != nil {
		return nil, err
	}
	for cur.Next(ctx) {
		mn := &MaintenanceModel{}
		err = cur.Decode(mn)
		if err != nil {
			return nil, err
		}
		result = append(result, mn)
	}

	return result, err
}
